@@ -12,8 +12,7 @@
 # Descr =
 # Descr =  Target: The directory in which the new file
 # Descr =          shall be created.
-# ButtonBgColor = #8AB
-# ButtonFgColor = #000
+# Icon = new_file.tga
 # Hotkey = Ctrl+F
 #[[END PROPERTIES]]
 */
@@ -49,8 +48,8 @@ if (IsExistingPath(newPath)) {
 }
 
 BatBegin("New file");
-BatWriteLineEchoed("<nul (set /p X=) > " + QuoteArg(newPath));
-BatWriteLine("if exist " + QuoteArg(newPath) + " (");
+BatWriteLineEchoed("<nul (set /p X=) > " + BatQuoteArg(newPath));
+BatWriteLine("if exist " + BatQuoteArg(newPath) + " (");
 BatWriteSendSelectKS([newPath]);
 BatWriteLine(") else (");
 BatWriteSetErrored();
