@@ -1,7 +1,7 @@
 //------------------------------------------------------------------------------
 // emInput.cpp
 //
-// Copyright (C) 2005-2010 Oliver Hamann.
+// Copyright (C) 2005-2012,2014-2018 Oliver Hamann.
 //
 // Homepage: http://eaglemode.sourceforge.net/
 //
@@ -19,7 +19,6 @@
 //------------------------------------------------------------------------------
 
 #include <emCore/emInput.h>
-#include <emCore/emThread.h>
 
 
 //==============================================================================
@@ -32,103 +31,112 @@ struct emInputKeyName {
 };
 
 static const emInputKeyName emInputKeyNames[] = {
-	{ EM_KEY_NONE         , "None"           },
-	{ EM_KEY_LEFT_BUTTON  , "Left-Button  "  },
-	{ EM_KEY_MIDDLE_BUTTON, "Middle-Button"  },
-	{ EM_KEY_RIGHT_BUTTON , "Right-Button "  },
-	{ EM_KEY_WHEEL_UP     , "Wheel-Up"       },
-	{ EM_KEY_WHEEL_DOWN   , "Wheel-Down"     },
-	{ EM_KEY_SHIFT        , "Shift"          },
-	{ EM_KEY_CTRL         , "Ctrl"           },
-	{ EM_KEY_ALT          , "Alt"            },
-	{ EM_KEY_META         , "Meta"           },
-	{ EM_KEY_ALT_GR       , "Alt-Gr"         },
-	{ EM_KEY_CURSOR_UP    , "Cursor-Up"      },
-	{ EM_KEY_CURSOR_DOWN  , "Cursor-Down"    },
-	{ EM_KEY_CURSOR_LEFT  , "Cursor-Left"    },
-	{ EM_KEY_CURSOR_RIGHT , "Cursor-Right"   },
-	{ EM_KEY_PAGE_UP      , "Page-Up"        },
-	{ EM_KEY_PAGE_DOWN    , "Page-Down"      },
-	{ EM_KEY_HOME         , "Home"           },
-	{ EM_KEY_END          , "End"            },
-	{ EM_KEY_PRINT        , "Print"          },
-	{ EM_KEY_PAUSE        , "Pause"          },
-	{ EM_KEY_MENU         , "Menu"           },
-	{ EM_KEY_INSERT       , "Insert"         },
-	{ EM_KEY_DELETE       , "Delete"         },
-	{ EM_KEY_BACKSPACE    , "Backspace"      },
-	{ EM_KEY_TAB          , "Tab"            },
-	{ EM_KEY_ENTER        , "Enter"          },
-	{ EM_KEY_ESCAPE       , "Escape"         },
-	{ EM_KEY_SPACE        , "Space"          },
-	{ EM_KEY_0            , "0"              },
-	{ EM_KEY_1            , "1"              },
-	{ EM_KEY_2            , "2"              },
-	{ EM_KEY_3            , "3"              },
-	{ EM_KEY_4            , "4"              },
-	{ EM_KEY_5            , "5"              },
-	{ EM_KEY_6            , "6"              },
-	{ EM_KEY_7            , "7"              },
-	{ EM_KEY_8            , "8"              },
-	{ EM_KEY_9            , "9"              },
-	{ EM_KEY_A            , "A"              },
-	{ EM_KEY_B            , "B"              },
-	{ EM_KEY_C            , "C"              },
-	{ EM_KEY_D            , "D"              },
-	{ EM_KEY_E            , "E"              },
-	{ EM_KEY_F            , "F"              },
-	{ EM_KEY_G            , "G"              },
-	{ EM_KEY_H            , "H"              },
-	{ EM_KEY_I            , "I"              },
-	{ EM_KEY_J            , "J"              },
-	{ EM_KEY_K            , "K"              },
-	{ EM_KEY_L            , "L"              },
-	{ EM_KEY_M            , "M"              },
-	{ EM_KEY_N            , "N"              },
-	{ EM_KEY_O            , "O"              },
-	{ EM_KEY_P            , "P"              },
-	{ EM_KEY_Q            , "Q"              },
-	{ EM_KEY_R            , "R"              },
-	{ EM_KEY_S            , "S"              },
-	{ EM_KEY_T            , "T"              },
-	{ EM_KEY_U            , "U"              },
-	{ EM_KEY_V            , "V"              },
-	{ EM_KEY_W            , "W"              },
-	{ EM_KEY_X            , "X"              },
-	{ EM_KEY_Y            , "Y"              },
-	{ EM_KEY_Z            , "Z"              },
-	{ EM_KEY_F1           , "F1"             },
-	{ EM_KEY_F2           , "F2"             },
-	{ EM_KEY_F3           , "F3"             },
-	{ EM_KEY_F4           , "F4"             },
-	{ EM_KEY_F5           , "F5"             },
-	{ EM_KEY_F6           , "F6"             },
-	{ EM_KEY_F7           , "F7"             },
-	{ EM_KEY_F8           , "F8"             },
-	{ EM_KEY_F9           , "F9"             },
-	{ EM_KEY_F10          , "F10"            },
-	{ EM_KEY_F11          , "F11"            },
-	{ EM_KEY_F12          , "F12"            }
+	{ EM_KEY_NONE          , "None"           },
+	{ EM_KEY_LEFT_BUTTON   , "Left-Button"    },
+	{ EM_KEY_MIDDLE_BUTTON , "Middle-Button"  },
+	{ EM_KEY_RIGHT_BUTTON  , "Right-Button"   },
+	{ EM_KEY_WHEEL_UP      , "Wheel-Up"       },
+	{ EM_KEY_WHEEL_DOWN    , "Wheel-Down"     },
+	{ EM_KEY_WHEEL_LEFT    , "Wheel-Left"     },
+	{ EM_KEY_WHEEL_RIGHT   , "Wheel-Right"    },
+	{ EM_KEY_BACK_BUTTON   , "Back-Button"    },
+	{ EM_KEY_FORWARD_BUTTON, "Forward-Button" },
+	{ EM_KEY_TOUCH         , "Touch"          },
+	{ EM_KEY_SHIFT         , "Shift"          },
+	{ EM_KEY_CTRL          , "Ctrl"           },
+	{ EM_KEY_ALT           , "Alt"            },
+	{ EM_KEY_META          , "Meta"           },
+	{ EM_KEY_ALT_GR        , "Alt-Gr"         },
+	{ EM_KEY_CURSOR_UP     , "Cursor-Up"      },
+	{ EM_KEY_CURSOR_DOWN   , "Cursor-Down"    },
+	{ EM_KEY_CURSOR_LEFT   , "Cursor-Left"    },
+	{ EM_KEY_CURSOR_RIGHT  , "Cursor-Right"   },
+	{ EM_KEY_PAGE_UP       , "Page-Up"        },
+	{ EM_KEY_PAGE_DOWN     , "Page-Down"      },
+	{ EM_KEY_HOME          , "Home"           },
+	{ EM_KEY_END           , "End"            },
+	{ EM_KEY_PRINT         , "Print"          },
+	{ EM_KEY_PAUSE         , "Pause"          },
+	{ EM_KEY_MENU          , "Menu"           },
+	{ EM_KEY_INSERT        , "Insert"         },
+	{ EM_KEY_DELETE        , "Delete"         },
+	{ EM_KEY_BACKSPACE     , "Backspace"      },
+	{ EM_KEY_TAB           , "Tab"            },
+	{ EM_KEY_ENTER         , "Enter"          },
+	{ EM_KEY_ESCAPE        , "Escape"         },
+	{ EM_KEY_SPACE         , "Space"          },
+	{ EM_KEY_0             , "0"              },
+	{ EM_KEY_1             , "1"              },
+	{ EM_KEY_2             , "2"              },
+	{ EM_KEY_3             , "3"              },
+	{ EM_KEY_4             , "4"              },
+	{ EM_KEY_5             , "5"              },
+	{ EM_KEY_6             , "6"              },
+	{ EM_KEY_7             , "7"              },
+	{ EM_KEY_8             , "8"              },
+	{ EM_KEY_9             , "9"              },
+	{ EM_KEY_A             , "A"              },
+	{ EM_KEY_B             , "B"              },
+	{ EM_KEY_C             , "C"              },
+	{ EM_KEY_D             , "D"              },
+	{ EM_KEY_E             , "E"              },
+	{ EM_KEY_F             , "F"              },
+	{ EM_KEY_G             , "G"              },
+	{ EM_KEY_H             , "H"              },
+	{ EM_KEY_I             , "I"              },
+	{ EM_KEY_J             , "J"              },
+	{ EM_KEY_K             , "K"              },
+	{ EM_KEY_L             , "L"              },
+	{ EM_KEY_M             , "M"              },
+	{ EM_KEY_N             , "N"              },
+	{ EM_KEY_O             , "O"              },
+	{ EM_KEY_P             , "P"              },
+	{ EM_KEY_Q             , "Q"              },
+	{ EM_KEY_R             , "R"              },
+	{ EM_KEY_S             , "S"              },
+	{ EM_KEY_T             , "T"              },
+	{ EM_KEY_U             , "U"              },
+	{ EM_KEY_V             , "V"              },
+	{ EM_KEY_W             , "W"              },
+	{ EM_KEY_X             , "X"              },
+	{ EM_KEY_Y             , "Y"              },
+	{ EM_KEY_Z             , "Z"              },
+	{ EM_KEY_F1            , "F1"             },
+	{ EM_KEY_F2            , "F2"             },
+	{ EM_KEY_F3            , "F3"             },
+	{ EM_KEY_F4            , "F4"             },
+	{ EM_KEY_F5            , "F5"             },
+	{ EM_KEY_F6            , "F6"             },
+	{ EM_KEY_F7            , "F7"             },
+	{ EM_KEY_F8            , "F8"             },
+	{ EM_KEY_F9            , "F9"             },
+	{ EM_KEY_F10           , "F10"            },
+	{ EM_KEY_F11           , "F11"            },
+	{ EM_KEY_F12           , "F12"            }
 };
 
 
 const char * emInputKeyToString(emInputKey key)
 {
-	static emThreadInitMutex initMutex;
-	static const char * table[256];
-	int i;
+	static const struct KeyToStringTable {
 
-	if (initMutex.Begin()) {
-		for (i=0; i<256; i++) table[i]=NULL;
-		for (i=0; i<(int)(sizeof(emInputKeyNames)/sizeof(emInputKeyName)); i++) {
-			if ((emInputKeyNames[i].Key&~255)==0) {
-				table[emInputKeyNames[i].Key]=emInputKeyNames[i].Name;
+		const char * table[256];
+
+		KeyToStringTable()
+		{
+			int i;
+			for (i=0; i<256; i++) table[i]=NULL;
+			for (i=0; i<(int)(sizeof(emInputKeyNames)/sizeof(emInputKeyName)); i++) {
+				if ((emInputKeyNames[i].Key&~255)==0) {
+					table[emInputKeyNames[i].Key]=emInputKeyNames[i].Name;
+				}
 			}
 		}
-		initMutex.End();
-	}
+
+	} keyToStringTable;
+
 	if ((key&~255)!=0) return NULL;
-	return table[key];
+	return keyToStringTable.table[key];
 }
 
 
@@ -150,29 +158,33 @@ static int emCompareInputKeyNames(
 
 emInputKey emStringToInputKey(const char * str)
 {
-	static emThreadInitMutex initMutex;
-	static emInputKeyName table[sizeof(emInputKeyNames)/sizeof(emInputKeyName)];
+	static const struct StringToKeyTable {
+
+		emInputKeyName table[sizeof(emInputKeyNames)/sizeof(emInputKeyName)];
+
+		StringToKeyTable()
+		{
+			memcpy(table,emInputKeyNames,sizeof(emInputKeyNames));
+			emSortArray(
+				table,
+				(int)(sizeof(emInputKeyNames)/sizeof(emInputKeyName)),
+				emCompareInputKeyNames,
+				(void*)NULL
+			);
+		}
+
+	} stringToKeyTable;
+
 	int i;
 
-	if (initMutex.Begin()) {
-		memcpy(table,emInputKeyNames,sizeof(emInputKeyNames));
-		emSortArray(
-			table,
-			(int)(sizeof(emInputKeyNames)/sizeof(emInputKeyName)),
-			emCompareInputKeyNames,
-			(void*)NULL
-		);
-		initMutex.End();
-	}
-
 	i=emBinarySearch(
-		table,
+		stringToKeyTable.table,
 		(int)(sizeof(emInputKeyNames)/sizeof(emInputKeyName)),
 		str,
 		emCompareInputKeyNameToName,
 		(void*)NULL
 	);
-	return i>=0 ? table[i].Key : EM_KEY_NONE;
+	return i>=0 ? stringToKeyTable.table[i].Key : EM_KEY_NONE;
 }
 
 
@@ -202,20 +214,14 @@ void emInputEvent::Setup(
 void emInputEvent::Eat()
 {
 	Key=EM_KEY_NONE;
-	Chars.Empty();
+	Chars.Clear();
 	Repeat=0;
 	Variant=0;
 }
 
 
-bool emInputEvent::IsKeyboardEvent() const
-{
-	return !IsEmpty() && !IsMouseEvent();
-}
-
-
 //==============================================================================
-//=============================== emInputState ================================
+//================================ emInputState ================================
 //==============================================================================
 
 emInputState::emInputState()
@@ -223,6 +229,7 @@ emInputState::emInputState()
 	MouseX=0.0;
 	MouseY=0.0;
 	memset(KeyStates,0,sizeof(KeyStates));
+	Touches.SetTuningLevel(4);
 }
 
 
@@ -231,6 +238,7 @@ emInputState::emInputState(const emInputState & inputState)
 	MouseX=inputState.MouseX;
 	MouseY=inputState.MouseY;
 	memcpy(KeyStates,inputState.KeyStates,sizeof(KeyStates));
+	Touches=inputState.Touches;
 }
 
 
@@ -239,27 +247,72 @@ emInputState & emInputState::operator = (const emInputState & inputState)
 	MouseX=inputState.MouseX;
 	MouseY=inputState.MouseY;
 	memcpy(KeyStates,inputState.KeyStates,sizeof(KeyStates));
+	Touches=inputState.Touches;
 	return *this;
 }
 
 
 bool emInputState::operator == (const emInputState & inputState) const
 {
-	return
-		MouseX==inputState.MouseX &&
-		MouseY==inputState.MouseY &&
-		memcmp(KeyStates,inputState.KeyStates,sizeof(KeyStates))==0
-	;
+	int i;
+
+	if (MouseX!=inputState.MouseX) return false;
+	if (MouseY!=inputState.MouseY) return false;
+	if (Touches.GetCount()!=inputState.Touches.GetCount()) return false;
+	for (i=Touches.GetCount()-1; i>=0; i--) {
+		if (Touches[i].Id!=inputState.Touches[i].Id) return false;
+		if (Touches[i].X!=inputState.Touches[i].X) return false;
+		if (Touches[i].Y!=inputState.Touches[i].Y) return false;
+	}
+	if (memcmp(KeyStates,inputState.KeyStates,sizeof(KeyStates))!=0) return false;
+	return true;
 }
 
 
 bool emInputState::operator != (const emInputState & inputState) const
 {
-	return
-		MouseX!=inputState.MouseX ||
-		MouseY!=inputState.MouseY ||
-		memcmp(KeyStates,inputState.KeyStates,sizeof(KeyStates))!=0
-	;
+	return !(*this == inputState);
+}
+
+
+int emInputState::SearchTouch(emUInt64 id) const
+{
+	int i;
+
+	for (i=Touches.GetCount()-1; i>=0; i--) {
+		if (Touches[i].Id==id) break;
+	}
+	return i;
+}
+
+
+void emInputState::AddTouch(emUInt64 id, double x, double y)
+{
+	Touches.AddNew();
+	SetTouch(Touches.GetCount()-1,id,x,y);
+}
+
+
+void emInputState::SetTouch(int index, emUInt64 id, double x, double y)
+{
+	Touch * t;
+
+	t=&Touches.GetWritable(index);
+	t->Id=id;
+	t->X=x;
+	t->Y=y;
+}
+
+
+void emInputState::RemoveTouch(int index)
+{
+	Touches.Remove(index);
+}
+
+
+void emInputState::ClearTouches()
+{
+	Touches.Clear();
 }
 
 
@@ -550,7 +603,7 @@ emInputHotkey::emInputHotkey(
 }
 
 
-void emInputHotkey::TryParse(const char * str) throw(emString)
+void emInputHotkey::TryParse(const char * str)
 {
 	char tmp[256];
 	const char * p;
@@ -578,7 +631,7 @@ void emInputHotkey::TryParse(const char * str) throw(emString)
 
 L_ERROR:
 	Packed=0;
-	throw emString::Format("Not a valid hotkey: %s",str);
+	throw emException("Not a valid hotkey: %s",str);
 }
 
 
