@@ -1,7 +1,7 @@
 //------------------------------------------------------------------------------
 // emPanel.cpp
 //
-// Copyright (C) 2004-2008 Oliver Hamann.
+// Copyright (C) 2004-2008,2011,2014-2017 Oliver Hamann.
 //
 // Homepage: http://eaglemode.sourceforge.net/
 //
@@ -54,7 +54,6 @@ emPanel::emPanel(ParentArg parent, const emString & name)
 		ClipX2=0;
 		ClipY2=0;
 		AEThresholdValue=EM_PANEL_DEFAULT_AE_THRESHOLD;
-		CreationNumber=++View.PanelCreationNumber;
 		CanvasColor=0;
 		PendingNoticeFlags=0;
 		Viewed=0;
@@ -62,18 +61,17 @@ emPanel::emPanel(ParentArg parent, const emString & name)
 		EnableSwitch=1;
 		Enabled=Parent->Enabled;
 		Focusable=1;
-		Visited=0;
-		InVisitedPath=0;
 		Active=0;
 		InActivePath=0;
 		PendingInput=0;
 		ChildrenLayoutInvalid=0;
-		AEEnabled=0;
+		AEInvalid=0;
 		AEDecisionInvalid=0;
 		AECalling=0;
 		AEExpanded=0;
 		CreatedByAE=Parent->AECalling;
 		AEThresholdType=VCT_AREA;
+		AutoplayHandling=APH_ITEM;
 		Parent->AvlInsertChild(this);
 		Parent->AddPendingNotice(NF_CHILD_LIST_CHANGED);
 		AddPendingNotice(
@@ -81,7 +79,6 @@ emPanel::emPanel(ParentArg parent, const emString & name)
 			NF_LAYOUT_CHANGED |
 			NF_VIEWING_CHANGED |
 			NF_ENABLE_CHANGED |
-			NF_VISIT_CHANGED |
 			NF_ACTIVE_CHANGED |
 			NF_FOCUS_CHANGED |
 			NF_VIEW_FOCUS_CHANGED |
@@ -100,7 +97,6 @@ emPanel::emPanel(ParentArg parent, const emString & name)
 		View.SupremeViewedPanel=this;
 		View.MinSVP=this;
 		View.MaxSVP=this;
-		View.VisitedPanel=this;
 		View.ActivePanel=this;
 		AvlTree=NULL;
 		Parent=NULL;
@@ -121,9 +117,8 @@ emPanel::emPanel(ParentArg parent, const emString & name)
 		ClipX1=ViewedX;
 		ClipY1=ViewedY;
 		ClipX2=ViewedX+ViewedWidth;
-		ClipY2=ViewedX+ViewedHeight;
+		ClipY2=ViewedY+ViewedHeight;
 		AEThresholdValue=EM_PANEL_DEFAULT_AE_THRESHOLD;
-		CreationNumber=++View.PanelCreationNumber;
 		CanvasColor=0;
 		PendingNoticeFlags=0;
 		Viewed=1;
@@ -131,25 +126,23 @@ emPanel::emPanel(ParentArg parent, const emString & name)
 		EnableSwitch=1;
 		Enabled=1;
 		Focusable=1;
-		Visited=1;
-		InVisitedPath=1;
 		Active=1;
 		InActivePath=1;
 		PendingInput=0;
 		ChildrenLayoutInvalid=0;
-		AEEnabled=0;
+		AEInvalid=0;
 		AEDecisionInvalid=0;
 		AECalling=0;
 		AEExpanded=0;
 		CreatedByAE=0;
 		AEThresholdType=VCT_AREA;
+		AutoplayHandling=APH_ITEM;
 		InvalidatePainting();
 		AddPendingNotice(
 			NF_CHILD_LIST_CHANGED |
 			NF_LAYOUT_CHANGED |
 			NF_VIEWING_CHANGED |
 			NF_ENABLE_CHANGED |
-			NF_VISIT_CHANGED |
 			NF_ACTIVE_CHANGED |
 			NF_FOCUS_CHANGED |
 			NF_VIEW_FOCUS_CHANGED |
@@ -170,39 +163,40 @@ emPanel::~emPanel()
 	if (View.SeekPosPanel==this) View.SetSeekPos(NULL,NULL);
 	DeleteAllChildren();
 	if (!Parent) {
-		if (View.IsPoppedUp()) View.ZoomOut();
+		if (View.IsPoppedUp()) View.RawZoomOut();
 		View.RootPanel=NULL;
 		View.SupremeViewedPanel=NULL;
 		View.MinSVP=NULL;
 		View.MaxSVP=NULL;
-		View.VisitedPanel=NULL;
 		View.ActivePanel=NULL;
-		View.VisitAdherent=false;
+		View.ActivationAdherent=false;
 		View.TitleInvalid=true;
 		View.CursorInvalid=true;
 		View.UpdateEngine->WakeUp();
 	}
 	else {
-		if (InVisitedPath) {
-			if (Parent->Viewed) {
-				Focusable=true;
-				View.VisitImmobile(Parent,false);
-			}
-			else {
+		if (InActivePath || View.SupremeViewedPanel==this) {
+			SetFocusable(false);
+			if (View.SupremeViewedPanel==this) {
 				LayoutX=-2.0;
 				LayoutY=-2.0;
 				LayoutWidth=1.0;
 				LayoutHeight=1.0;
 				CanvasColor=0;
-				Focusable=true;
-				View.ProtectSeeking++;
-				View.VisitFullsized(Parent,!InActivePath && View.VisitAdherent);
-				View.ProtectSeeking--;
+				if ((View.GetViewFlags()&emView::VF_POPUP_ZOOM)!=0 && !View.IsPoppedUp()) {
+					View.RawZoomOut();
+				}
+				else {
+					View.RawVisitFullsized(Parent);
+				}
 			}
-			if (InVisitedPath) {
-				emFatalError("emPanel::~emPanel: Could not to get rid of the visit.");
+			if (InActivePath || View.SupremeViewedPanel==this) {
+				emFatalError("emPanel::~emPanel: Could not to get rid of activation or SVP status.");
 			}
 		}
+		if (View.MinSVP==this) {
+			View.MinSVP=Parent;
+		}
 		View.RestartInputRecursion=true;
 		if (InViewedPath) {
 			View.SVPChoiceInvalid=true;
@@ -225,9 +219,6 @@ emPanel::~emPanel()
 		NoticeNode.Next=NULL;
 		NoticeNode.Prev=NULL;
 	}
-	if (View.ActivationCandidate==this) {
-		View.SetActivationCandidate(NULL);
-	}
 }
 
 
@@ -328,14 +319,21 @@ emArray<emString> emPanel::DecodeIdentity(const char * identity)
 }
 
 
-emString emPanel::GetTitle()
+emString emPanel::GetTitle() const
 {
 	if (Parent) return Parent->GetTitle();
 	else return "untitled";
 }
 
 
-emPanel * emPanel::GetChild(const char * name)
+emString emPanel::GetIconFileName() const
+{
+	if (Parent) return Parent->GetIconFileName();
+	else return emString();
+}
+
+
+emPanel * emPanel::GetChild(const char * name) const
 {
 	EM_AVL_SEARCH_VARS(emPanel)
 	int d;
@@ -484,8 +482,8 @@ void emPanel::Layout(
 )
 {
 	emPanel * p;
-	double x1,y1,x2,y2,rx,ry,ra,ra2;
-	bool adherent,zoomedOut;
+	double x1,y1,x2,y2,rx,ry,ra;
+	bool zoomedOut;
 
 	if (LayoutWidth<1E-100) LayoutWidth=1E-100;
 	if (LayoutHeight<1E-100) LayoutHeight=1E-100;
@@ -525,7 +523,7 @@ void emPanel::Layout(
 
 	if (!Parent) {
 		zoomedOut=View.IsZoomedOut();
-		p=View.GetVisitedPanel(&rx,&ry,&ra,&adherent);
+		p=View.GetVisitedPanel(&rx,&ry,&ra);
 		LayoutX=layoutX;
 		LayoutY=layoutY;
 		LayoutWidth=layoutWidth;
@@ -533,24 +531,26 @@ void emPanel::Layout(
 		CanvasColor=canvasColor;
 		if (!View.SettingGeometry) {
 			if (zoomedOut) {
-				ra=View.HomeWidth*GetHeight()/View.HomePixelTallness/View.HomeHeight;
-				ra2=View.HomeHeight/GetHeight()*View.HomePixelTallness/View.HomeWidth;
-				if (ra<ra2) ra=ra2;
-				View.VisitRelBy(this,0.0,0.0,ra,true);
+				View.RawZoomOut(true);
 			}
 			else if (p) {
-				View.VisitRel(p,rx,ry,ra,adherent,true);
+				View.RawVisit(p,rx,ry,ra,true);
 			}
 		}
 	}
-	else if (InVisitedPath && !View.SettingGeometry && !View.IsZoomedOut()) {
-		p=View.GetVisitedPanel(&rx,&ry,&ra,&adherent);
+	else if (
+		InViewedPath &&
+		(InActivePath || !Parent->Viewed) &&
+		!View.SettingGeometry &&
+		!View.IsZoomedOut()
+	) {
+		p=View.GetVisitedPanel(&rx,&ry,&ra);
 		LayoutX=layoutX;
 		LayoutY=layoutY;
 		LayoutWidth=layoutWidth;
 		LayoutHeight=layoutHeight;
 		CanvasColor=canvasColor;
-		View.VisitRel(p,rx,ry,ra,adherent,true);
+		View.RawVisit(p,rx,ry,ra,true);
 	}
 	else if (Parent->Viewed) {
 		LayoutX=layoutX;
@@ -608,14 +608,45 @@ void emPanel::Layout(
 }
 
 
-void emPanel::GetEssenceRect(
-	double * pX, double * pY, double * pW, double * pH
-)
+void emPanel::GetSubstanceRect(
+	double * pX, double * pY, double * pW, double * pH, double * pR
+) const
 {
 	*pX=0.0;
 	*pY=0.0;
 	*pW=1.0;
 	*pH=GetHeight();
+	*pR=0.0;
+}
+
+
+bool emPanel::IsPointInSubstanceRect(double x, double y) const
+{
+	double sx,sy,sw,sh,sr,dx,dy,sw2,sh2;
+
+	if (x<0.0 || x>=1.0 || y<0.0 || y>=GetHeight()) return false;
+	GetSubstanceRect(&sx,&sy,&sw,&sh,&sr);
+	sw2=sw*0.5;
+	dx=fabs(x-sx-sw2);
+	if (dx>sw2) return false;
+	sh2=sh*0.5;
+	dy=fabs(y-sy-sh2);
+	if (dy>sh2) return false;
+	if (sr>sw2) sr=sw2;
+	if (sr>sh2) sr=sh2;
+	dx-=sw2-sr;
+	dy-=sh2-sr;
+	return dx<0.0 || dy<0.0 || dx*dx+dy*dy <= sr*sr;
+}
+
+
+void emPanel::GetEssenceRect(
+	double * pX, double * pY, double * pW, double * pH
+) const
+{
+	double r;
+
+	GetSubstanceRect(pX,pY,pW,pH,&r);
 }
 
 
@@ -706,42 +737,39 @@ void emPanel::SetEnableSwitch(bool enableSwitch)
 
 void emPanel::SetFocusable(bool focusable)
 {
-	if (Parent && ((bool)Focusable)!=focusable) {
-		Focusable=focusable;
-		if (Focusable) {
-			if (InVisitedPath && !InActivePath) {
-				View.VisitImmobile(View.ActivePanel,View.VisitAdherent);
-			}
+	if (((bool)Focusable)!=focusable) {
+		if (!Parent && !focusable) {
+			emDLog("emPanel::SetFocusable: a root panel cannot be set unfocusable");
+			return;
 		}
-		else {
-			if (Active) {
-				View.VisitImmobile(Parent,false);
-			}
+		Focusable=focusable;
+		if (!Focusable && Active) {
+			View.SetActivePanel(Parent,false);
 		}
 	}
 }
 
 
-emPanel * emPanel::GetFocusableParent()
+emPanel * emPanel::GetFocusableParent() const
 {
-	emPanel * p;
+	const emPanel * p;
 
 	p=this;
 	do {
 		p=p->Parent;
 	} while (p && !p->Focusable);
-	return p;
+	return (emPanel*)p;
 }
 
 
-emPanel * emPanel::GetFocusableFirstChild()
+emPanel * emPanel::GetFocusableFirstChild() const
 {
-	emPanel * p;
+	const emPanel * p;
 
 	p=FirstChild;
 	if (!p) return NULL;
 	for (;;) {
-		if (p->Focusable) return p;
+		if (p->Focusable) return (emPanel*)p;
 		if (p->FirstChild) p=p->FirstChild;
 		else {
 			while (!p->Next) {
@@ -754,14 +782,14 @@ emPanel * emPanel::GetFocusableFirstChild()
 }
 
 
-emPanel * emPanel::GetFocusableLastChild()
+emPanel * emPanel::GetFocusableLastChild() const
 {
-	emPanel * p;
+	const emPanel * p;
 
 	p=LastChild;
 	if (!p) return NULL;
 	for (;;) {
-		if (p->Focusable) return p;
+		if (p->Focusable) return (emPanel*)p;
 		if (p->LastChild) p=p->LastChild;
 		else {
 			while (!p->Prev) {
@@ -774,9 +802,9 @@ emPanel * emPanel::GetFocusableLastChild()
 }
 
 
-emPanel * emPanel::GetFocusablePrev()
+emPanel * emPanel::GetFocusablePrev() const
 {
-	emPanel * p;
+	const emPanel * p;
 
 	for (p=this;;) {
 		while (!p->Prev) {
@@ -785,7 +813,7 @@ emPanel * emPanel::GetFocusablePrev()
 		}
 		p=p->Prev;
 		for (;;) {
-			if (p->Focusable) return p;
+			if (p->Focusable) return (emPanel*)p;
 			if (!p->LastChild) break;
 			p=p->LastChild;
 		}
@@ -793,9 +821,9 @@ emPanel * emPanel::GetFocusablePrev()
 }
 
 
-emPanel * emPanel::GetFocusableNext()
+emPanel * emPanel::GetFocusableNext() const
 {
-	emPanel * p;
+	const emPanel * p;
 
 	for (p=this;;) {
 		while (!p->Next) {
@@ -804,7 +832,7 @@ emPanel * emPanel::GetFocusableNext()
 		}
 		p=p->Next;
 		for (;;) {
-			if (p->Focusable) return p;
+			if (p->Focusable) return (emPanel*)p;
 			if (!p->FirstChild) break;
 			p=p->FirstChild;
 		}
@@ -812,22 +840,16 @@ emPanel * emPanel::GetFocusableNext()
 }
 
 
-void emPanel::Activate()
+void emPanel::Activate(bool adherent)
 {
-	View.VisitLazy(this,true);
+	View.SetActivePanel(this,adherent);
 }
 
 
-void emPanel::ActivateLater()
-{
-	View.SetActivationCandidate(this);
-}
-
-
-void emPanel::Focus()
+void emPanel::Focus(bool adherent)
 {
 	View.Focus();
-	View.VisitLazy(this,true);
+	View.SetActivePanel(this,adherent);
 }
 
 
@@ -1019,6 +1041,39 @@ emUInt64 emPanel::GetMemoryLimit() const
 }
 
 
+double emPanel::GetTouchEventPriority(double touchX, double touchY) const
+{
+	return Focusable ? 1.0 : 0.0;
+}
+
+
+void emPanel::SetAutoplayHandling(AutoplayHandlingFlags flags)
+{
+	AutoplayHandling=flags;
+}
+
+
+bool emPanel::IsContentReady(bool * pReadying) const
+{
+	if (pReadying) *pReadying=false;
+	return IsAutoExpanded();
+}
+
+
+bool emPanel::GetPlaybackState(bool * pPlaying, double * pPos) const
+{
+	if (pPlaying) *pPlaying=false;
+	if (pPos) *pPos=0.0;
+	return false;
+}
+
+
+bool emPanel::SetPlaybackState(bool playing, double pos)
+{
+	return false;
+}
+
+
 bool emPanel::Cycle()
 {
 	return false;
@@ -1034,7 +1089,7 @@ void emPanel::Input(
 	emInputEvent & event, const emInputState & state, double mx, double my
 )
 {
-	if (Focusable && event.IsMouseEvent()) {
+	if (Focusable && (event.IsMouseEvent() || event.IsTouchEvent())) {
 		Focus();
 		event.Eat();
 	}
@@ -1092,7 +1147,11 @@ void emPanel::Input(
 				event.Eat();
 			}
 			else if (state.IsAltMod()) {
-				View.VisitFullsized(this,View.IsVisitAdherent());
+				View.VisitFullsized(this,View.IsActivationAdherent());
+				event.Eat();
+			}
+			else if (state.IsShiftAltMod()) {
+				View.VisitFullsized(this,View.IsActivationAdherent(),true);
 				event.Eat();
 			}
 			break;
@@ -1109,34 +1168,24 @@ void emPanel::Input(
 }
 
 
-emCursor emPanel::GetCursor()
+emCursor emPanel::GetCursor() const
 {
 	if (Parent) return Parent->GetCursor();
 	else return emCursor::NORMAL;
 }
 
 
-bool emPanel::IsOpaque()
+bool emPanel::IsOpaque() const
 {
 	return false;
 }
 
 
-void emPanel::Paint(const emPainter & painter, emColor canvasColor)
+void emPanel::Paint(const emPainter & painter, emColor canvasColor) const
 {
 }
 
 
-void emPanel::EnableAutoExpansion()
-{
-	if (!AEEnabled) {
-		AEEnabled=1;
-		AEDecisionInvalid=1;
-		if (!NoticeNode.Next) View.AddToNoticeList(&NoticeNode);
-	}
-}
-
-
 void emPanel::AutoExpand()
 {
 }
@@ -1173,7 +1222,7 @@ const char * emPanel::GetSoughtName() const
 }
 
 
-bool emPanel::IsHopeForSeeking()
+bool emPanel::IsHopeForSeeking() const
 {
 	return false;
 }
@@ -1231,11 +1280,9 @@ void emPanel::InvalidatePainting(double x, double y, double w, double h)
 
 void emPanel::InvalidateAutoExpansion()
 {
-	if (AEExpanded) {
-		AEExpanded=0;
-		AEDecisionInvalid=1;
+	if (!AEInvalid && AEExpanded) {
+		AEInvalid=1;
 		if (!NoticeNode.Next) View.AddToNoticeList(&NoticeNode);
-		AutoShrink();
 	}
 }
 
@@ -1246,24 +1293,94 @@ void emPanel::InvalidateControlPanel()
 }
 
 
+emString emPanel::GetTitle()
+{
+	return ((const emPanel*)this)->GetTitle();
+}
+
+
+emString emPanel::GetIconFileName()
+{
+	return ((const emPanel*)this)->GetIconFileName();
+}
+
+
+void emPanel::GetSubstanceRect(
+	double * pX, double * pY, double * pW, double * pH, double * pR
+)
+{
+	return ((const emPanel*)this)->GetSubstanceRect(pX,pY,pW,pH,pR);
+}
+
+
+void emPanel::GetEssenceRect(
+	double * pX, double * pY, double * pW, double * pH
+)
+{
+	return ((const emPanel*)this)->GetEssenceRect(pX,pY,pW,pH);
+}
+
+
+double emPanel::GetTouchEventPriority(double touchX, double touchY)
+{
+	return ((const emPanel*)this)->GetTouchEventPriority(touchX,touchY);
+}
+
+
+emCursor emPanel::GetCursor()
+{
+	return ((const emPanel*)this)->GetCursor();
+}
+
+
+bool emPanel::IsOpaque()
+{
+	return ((const emPanel*)this)->IsOpaque();
+}
+
+
+void emPanel::Paint(const emPainter & painter, emColor canvasColor)
+{
+	((const emPanel*)this)->Paint(painter,canvasColor);
+}
+
+
+bool emPanel::IsHopeForSeeking()
+{
+	return ((const emPanel*)this)->IsHopeForSeeking();
+}
+
+
 void emPanel::HandleNotice()
 {
 	NoticeFlags flags;
 
+	if (AEInvalid) {
+		AEInvalid=0;
+		if (AEExpanded) {
+			AEExpanded=0;
+			AEDecisionInvalid=1;
+			AutoShrink();
+		}
+	}
+
 	flags=PendingNoticeFlags;
 	if (flags) {
-		if (AEEnabled) {
-			if (flags&(NF_SOUGHT_NAME_CHANGED|NF_VIEWING_CHANGED)) {
-				AEDecisionInvalid=1;
+		if (flags&(NF_SOUGHT_NAME_CHANGED|NF_VIEWING_CHANGED)) {
+			if (
+				View.SeekPosPanel==this ||
+				GetViewCondition((ViewConditionType)AEThresholdType)>=AEThresholdValue
+			) {
+				if (!AEExpanded) AEDecisionInvalid=1;
 			}
-			if (AEDecisionInvalid) {
-				if (!NoticeNode.Next) View.AddToNoticeList(&NoticeNode);
+			else {
+				if (AEExpanded) AEDecisionInvalid=1;
 			}
 		}
 		if (flags&(NF_LAYOUT_CHANGED|NF_CHILD_LIST_CHANGED)) {
 			if (FirstChild) ChildrenLayoutInvalid=1;
 		}
-		if (ChildrenLayoutInvalid) {
+		if (AEDecisionInvalid || ChildrenLayoutInvalid) {
 			if (!NoticeNode.Next) View.AddToNoticeList(&NoticeNode);
 		}
 		PendingNoticeFlags=0;
@@ -1273,25 +1390,23 @@ void emPanel::HandleNotice()
 
 	if (AEDecisionInvalid) {
 		AEDecisionInvalid=0;
-		if (AEEnabled) {
-			if (
-				GetSoughtName() ||
-				GetViewCondition((ViewConditionType)AEThresholdType)>=AEThresholdValue
-			) {
-				if (!AEExpanded) {
-					AEExpanded=1;
-					AECalling=1;
-					AutoExpand();
-					AECalling=0;
-					if (PendingNoticeFlags) return;
-				}
+		if (
+			View.SeekPosPanel==this ||
+			GetViewCondition((ViewConditionType)AEThresholdType)>=AEThresholdValue
+		) {
+			if (!AEExpanded) {
+				AEExpanded=1;
+				AECalling=1;
+				AutoExpand();
+				AECalling=0;
+				if (PendingNoticeFlags) return;
 			}
-			else {
-				if (AEExpanded) {
-					AEExpanded=0;
-					AutoShrink();
-					if (PendingNoticeFlags) return;
-				}
+		}
+		else {
+			if (AEExpanded) {
+				AEExpanded=0;
+				AutoShrink();
+				if (PendingNoticeFlags) return;
 			}
 		}
 	}
