@@ -1,7 +1,7 @@
 //------------------------------------------------------------------------------
 // emDirEntryPanel.cpp
 //
-// Copyright (C) 2004-2011 Oliver Hamann.
+// Copyright (C) 2004-2012,2014,2016-2017 Oliver Hamann.
 //
 // Homepage: http://eaglemode.sourceforge.net/
 //
@@ -18,18 +18,8 @@
 // along with this program. If not, see <http://www.gnu.org/licenses/>.
 //------------------------------------------------------------------------------
 
-#include <sys/stat.h>
 #if defined(_WIN32)
 #	include <windows.h>
-#	ifndef S_IXUSR
-#		define S_IXUSR 0100
-#	endif
-#	ifndef S_IXGRP
-#		define S_IXGRP 0010
-#	endif
-#	ifndef S_IXOTH
-#		define S_IXOTH 0001
-#	endif
 #else
 #	include <unistd.h>
 #endif
@@ -39,6 +29,17 @@
 #include <emFileMan/emDirEntryAltPanel.h>
 #include <emFileMan/emFileManControlPanel.h>
 #include <emCore/emRes.h>
+#if defined(_WIN32)
+#	ifndef S_IXUSR
+#		define S_IXUSR 0100
+#	endif
+#	ifndef S_IXGRP
+#		define S_IXGRP 0010
+#	endif
+#	ifndef S_IXOTH
+#		define S_IXOTH 0001
+#	endif
+#endif
 
 
 emDirEntryPanel::emDirEntryPanel(
@@ -53,6 +54,8 @@ emDirEntryPanel::emDirEntryPanel(
 	AddWakeUpSignal(FileMan->GetSelectionSignal());
 	AddWakeUpSignal(Config->GetChangeSignal());
 
+	SetAutoplayHandling(APH_CUTOFF_AT_SUBITEMS);
+
 	UpdateBgColor();
 }
 
@@ -91,12 +94,58 @@ void emDirEntryPanel::UpdateDirEntry(const emDirEntry & dirEntry)
 }
 
 
-emString emDirEntryPanel::GetTitle()
+emString emDirEntryPanel::GetTitle() const
 {
 	return DirEntry.GetPath();
 }
 
 
+emString emDirEntryPanel::GetIconFileName() const
+{
+	static int recursive = 0; //??? Not good
+	emString icon;
+	emPanel * p;
+
+	if (recursive <= 0) {
+		p=GetChild(ContentName);
+		if (p) {
+			recursive++;
+			icon = p->GetIconFileName(); // May call the parent, i.e. this!
+			recursive--;
+			return icon;
+		}
+	}
+
+	if (DirEntry.IsDirectory()) return "directory.tga";
+	else return "file.tga";
+}
+
+
+void emDirEntryPanel::GetSubstanceRect(
+	double * pX, double * pY,
+	double * pW, double * pH,
+	double * pR
+) const
+{
+	const emFileManTheme * theme;
+
+	theme = &Config->GetTheme();
+	*pX=theme->BackgroundX;
+	*pY=theme->BackgroundY;
+	*pW=theme->BackgroundW;
+	*pH=theme->BackgroundH;
+	*pR=emMin(theme->BackgroundRX.Get(),theme->BackgroundRY.Get());
+}
+
+
+bool emDirEntryPanel::IsContentReady(bool * pReadying) const
+{
+	if (!emPanel::IsContentReady(pReadying)) return false;
+	if (pReadying) *pReadying=false;
+	return GetChild(ContentName)!=NULL;
+}
+
+
 bool emDirEntryPanel::Cycle()
 {
 	if (IsSignaled(FileMan->GetSelectionSignal())) {
@@ -114,7 +163,7 @@ bool emDirEntryPanel::Cycle()
 
 void emDirEntryPanel::Notice(NoticeFlags flags)
 {
-	if ((flags&(NF_VIEWING_CHANGED|NF_SOUGHT_NAME_CHANGED|NF_VISIT_CHANGED))!=0) {
+	if ((flags&(NF_VIEWING_CHANGED|NF_SOUGHT_NAME_CHANGED|NF_ACTIVE_CHANGED))!=0) {
 		UpdateContentPanel();
 		UpdateAltPanel();
 	}
@@ -125,10 +174,42 @@ void emDirEntryPanel::Input(
 	emInputEvent & event, const emInputState & state, double mx, double my
 )
 {
+	const emFileManTheme * theme;
+	double cx,cy,cw,ch;
+	emPanel * p;
+
 	if (event.IsKeyboardEvent() && !IsActive()) {
 		event.Eat();
 	}
 
+	if (event.IsMouseEvent()) {
+		theme = &Config->GetTheme();
+		if (DirEntry.IsDirectory()) {
+			cx=theme->DirContentX;
+			cy=theme->DirContentY;
+			cw=theme->DirContentW;
+			ch=theme->DirContentH;
+		}
+		else {
+			cx=theme->FileContentX;
+			cy=theme->FileContentY;
+			cw=theme->FileContentW;
+			ch=theme->FileContentH;
+		}
+		if (mx>=cx && mx<cx+cw && my>=cy && my<cy+ch) {
+			p=GetChild(ContentName);
+			if (p) {
+				if (!p->IsFocusable()) {
+					p=p->GetFocusableFirstChild();
+				}
+				if (p) {
+					p->Focus();
+					event.Eat();
+				}
+			}
+		}
+	}
+
 	switch (event.GetKey()) {
 	case EM_KEY_LEFT_BUTTON:
 		if (event.GetRepeat() && state.IsNoMod()) {
@@ -180,7 +261,7 @@ void emDirEntryPanel::Input(
 }
 
 
-bool emDirEntryPanel::IsOpaque()
+bool emDirEntryPanel::IsOpaque() const
 {
 	const emFileManTheme * theme;
 
@@ -198,10 +279,11 @@ bool emDirEntryPanel::IsOpaque()
 }
 
 
-void emDirEntryPanel::Paint(const emPainter & painter, emColor canvasColor)
+void emDirEntryPanel::Paint(const emPainter & painter, emColor canvasColor) const
 {
 	const emFileManTheme * theme;
 	emColor color;
+	emString str;
 	double t;
 
 	theme = &Config->GetTheme();
@@ -268,12 +350,20 @@ void emDirEntryPanel::Paint(const emPainter & painter, emColor canvasColor)
 #endif
 	else color=theme->OtherNameColor;
 	if (DirEntry.IsHidden()) color=color.GetTransparented(27.0F);
+	str=DirEntry.GetName();
+#if defined(_WIN32)
+	if (DirEntry.IsDrive() && !DirEntry.GetDriveName().IsEmpty()) {
+		str=emString::Format(
+			"%s - %s",DirEntry.GetName().Get(),DirEntry.GetDriveName().Get()
+		);
+	}
+#endif
 	painter.PaintTextBoxed(
 		theme->NameX,
 		theme->NameY,
 		theme->NameW,
 		theme->NameH,
-		DirEntry.GetName(),
+		str,
 		theme->NameH,
 		color,
 		canvasColor,
@@ -352,6 +442,18 @@ void emDirEntryPanel::Paint(const emPainter & painter, emColor canvasColor)
 				theme->FileInnerBorderImgB,
 				255,canvasColor,0757
 			);
+			if (
+				theme->FileContentX + 1E-10 <
+					theme->FileInnerBorderX + theme->FileInnerBorderL ||
+				theme->FileContentY + 1E-10 <
+					theme->FileInnerBorderY + theme->FileInnerBorderT ||
+				theme->FileContentX + theme->FileContentW - 1E-10 >
+					theme->FileInnerBorderX + theme->FileInnerBorderW - theme->FileInnerBorderR ||
+				theme->FileContentY + theme->FileContentH - 1E-10 >
+					theme->FileInnerBorderY + theme->FileInnerBorderH - theme->FileInnerBorderB
+			) {
+				canvasColor=0;
+			}
 			painter.PaintRect(
 				theme->FileContentX,
 				theme->FileContentY,
@@ -381,7 +483,7 @@ emPanel * emDirEntryPanel::CreateControlPanel(
 void emDirEntryPanel::PaintInfo(
 	const emPainter & painter, double infoX, double infoY,
 	double infoW, double infoH, emAlignment alignment, emColor canvasColor
-)
+) const
 {
 	const char * label[6] = {
 		"Type",
@@ -505,6 +607,22 @@ void emDirEntryPanel::PaintInfo(
 		);
 	}
 	else {
+#if defined(_WIN32)
+		if (DirEntry.IsDrive()) {
+			switch (DirEntry.GetDriveType()) {
+			case DRIVE_REMOVABLE: p="Removable Drive"; break;
+			case DRIVE_REMOTE:    p="Network Drive"; break;
+			case DRIVE_CDROM:     p="Optical Disk"; break;
+			case DRIVE_RAMDISK:   p="RAM Disk"; break;
+			default:              p="Disk Drive"; break;
+			}
+			if (!DirEntry.GetDriveFileSystem().IsEmpty()) {
+				snprintf(tmp,sizeof(tmp),"%s: %s",p,DirEntry.GetDriveFileSystem().Get());
+				tmp[sizeof(tmp)-1]=0;
+				p=tmp;
+			}
+		}
+#endif
 		painter.PaintTextBoxed(
 			bx[0],by[0],bw[0],bh[0],
 			p,bh[0],
@@ -564,6 +682,9 @@ void emDirEntryPanel::PaintInfo(
 		0.5,false
 	);
 
+#if defined(_WIN32)
+	if (!DirEntry.IsDirectory()) {
+#endif
 	len=emUInt64ToStr(tmp,sizeof(tmp),DirEntry.GetStat()->st_size);
 	cw=painter.GetTextSize("X",bh[4],false);
 	ws=bw[4]/(cw*len*16/15);
@@ -582,7 +703,13 @@ void emDirEntryPanel::PaintInfo(
 		}
 		x+=cw/5*ws;
 	}
+#if defined(_WIN32)
+	}
+#endif
 
+#if defined(_WIN32)
+	if (!DirEntry.IsDrive()) {
+#endif
 	FormatTime(DirEntry.GetStat()->st_mtime,tmp,bw[5]/bh[5]<6.0);
 	painter.PaintTextBoxed(
 		bx[5],by[5],bw[5],bh[5],
@@ -591,6 +718,9 @@ void emDirEntryPanel::PaintInfo(
 		EM_ALIGN_LEFT,EM_ALIGN_LEFT,
 		0.5,true
 	);
+#if defined(_WIN32)
+	}
+#endif
 }
 
 
@@ -651,7 +781,7 @@ void emDirEntryPanel::UpdateContentPanel(bool forceRecreation, bool forceRelayou
 			forceRelayout=true;
 		}
 	}
-	else if (p && !p->IsInVisitedPath()) {
+	else if (p && !p->IsInActivePath() && (!p->IsInViewedPath() || IsViewed())) {
 		delete p;
 		p=NULL;
 	}
@@ -694,7 +824,7 @@ void emDirEntryPanel::UpdateAltPanel(bool forceRecreation, bool forceRelayout)
 			forceRelayout=true;
 		}
 	}
-	else if (p && !p->IsInVisitedPath()) {
+	else if (p && !p->IsInActivePath() && (!p->IsInViewedPath() || IsViewed())) {
 		delete p;
 		p=NULL;
 	}
