@@ -1,7 +1,7 @@
 //------------------------------------------------------------------------------
 // emMainContentPanel.cpp
 //
-// Copyright (C) 2007-2008 Oliver Hamann.
+// Copyright (C) 2007-2008,2016-2017 Oliver Hamann.
 //
 // Homepage: http://eaglemode.sourceforge.net/
 //
@@ -28,6 +28,7 @@ emMainContentPanel::emMainContentPanel(ParentArg parent, const emString & name)
 {
 	SkyImage=emGetInsResImage(GetRootContext(),"emMain","Sky.tga",3);
 	new emVirtualCosmosPanel(this,"");
+	if (!GetParent()) SetAutoplayHandling(APH_DIRECTORY);
 	UpdateCoordinates();
 	UpdateChildLayout();
 }
@@ -38,12 +39,18 @@ emMainContentPanel::~emMainContentPanel()
 }
 
 
-emString emMainContentPanel::GetTitle()
+emString emMainContentPanel::GetTitle() const
 {
 	return "Eagle";
 }
 
 
+emString emMainContentPanel::GetIconFileName() const
+{
+	return "virtual_cosmos.tga";
+}
+
+
 void emMainContentPanel::Layout(
 	double layoutX, double layoutY, double layoutWidth, double layoutHeight,
 	emColor canvasColor
@@ -63,13 +70,13 @@ void emMainContentPanel::Layout(
 }
 
 
-bool emMainContentPanel::IsOpaque()
+bool emMainContentPanel::IsOpaque() const
 {
 	return true;
 }
 
 
-void emMainContentPanel::Paint(const emPainter & painter, emColor canvasColor)
+void emMainContentPanel::Paint(const emPainter & painter, emColor canvasColor) const
 {
 	painter.PaintImage(0,0,1,GetHeight(),SkyImage,255,canvasColor);
 	PaintEagle(
