@@ -1,7 +1,7 @@
 //------------------------------------------------------------------------------
 // emPsRenderer.cpp
 //
-// Copyright (C) 2006-2010 Oliver Hamann.
+// Copyright (C) 2006-2011,2014,2017-2018 Oliver Hamann.
 //
 // Homepage: http://eaglemode.sourceforge.net/
 //
@@ -19,6 +19,7 @@
 //------------------------------------------------------------------------------
 
 #include <emPs/emPsRenderer.h>
+#include <emCore/emInstallInfo.h>
 
 
 emRef<emPsRenderer> emPsRenderer::Acquire(emRootContext & rootContext)
@@ -114,7 +115,7 @@ bool emPsRenderer::Cycle()
 	switch (MainState) {
 
 L_ENTER_COLD_WAIT_JOB:
-		CurrentDocument.Empty();
+		CurrentDocument.Clear();
 		PSAgent.ReleaseAccess();
 		MainState=COLD_WAIT_JOB;
 	case COLD_WAIT_JOB:
@@ -138,8 +139,8 @@ L_ENTER_PREPARE_PROCESS:
 		try {
 			TryStartProcess();
 		}
-		catch (emString errorMessage) {
-			FailAllJobs(errorMessage);
+		catch (emException & exception) {
+			FailAllJobs(exception.GetText());
 			goto L_ENTER_COLD_WAIT_JOB;
 		}
 		PrepareWritingStartup();
@@ -155,8 +156,8 @@ L_ENTER_PREPARE_PROCESS:
 			TryRead();
 			TryWrite();
 		}
-		catch (emString errorMessage) {
-			FailDocJobs(errorMessage);
+		catch (emException & exception) {
+			FailDocJobs(exception.GetText());
 			goto L_ENTER_QUIT_PROCESS;
 		}
 		if (IsReadingFinished()) goto L_ENTER_RUN_JOB;
@@ -168,6 +169,7 @@ L_ENTER_PREPARE_PROCESS:
 		break;
 
 L_ENTER_RUN_JOB:
+		if (CurrentDocument.GetDataRefCount()<=1) goto L_ENTER_QUIT_PROCESS;
 		job=SearchBestSameDocJob();
 		if (!job) goto L_ENTER_QUIT_PROCESS;
 		SetJobState(job,JS_RUNNING);
@@ -190,8 +192,8 @@ L_ENTER_RUN_JOB:
 				readProceeded=TryRead();
 				writeProceeded=TryWrite();
 			}
-			catch (emString errorMessage) {
-				FailDocJobs(errorMessage);
+			catch (emException & exception) {
+				FailDocJobs(exception.GetText());
 				goto L_ENTER_QUIT_PROCESS;
 			}
 			if (IsReadingFinished()) {
@@ -213,8 +215,10 @@ L_ENTER_HOT_WAIT_JOB:
 		Timer.Start(3000);
 		MainState=HOT_WAIT_JOB;
 	case HOT_WAIT_JOB:
+		if (CurrentDocument.GetDataRefCount()<=1) goto L_ENTER_QUIT_PROCESS;
 		if (FirstJob) goto L_ENTER_HOT_WAIT_ACCESS;
 		if (!Timer.IsRunning()) goto L_ENTER_QUIT_PROCESS;
+		busy=true;
 		break;
 
 L_ENTER_HOT_WAIT_ACCESS:
@@ -222,13 +226,15 @@ L_ENTER_HOT_WAIT_ACCESS:
 		PSAgent.RequestAccess();
 		MainState=HOT_WAIT_ACCESS;
 	case HOT_WAIT_ACCESS:
+		if (CurrentDocument.GetDataRefCount()<=1) goto L_ENTER_QUIT_PROCESS;
 		if (!FirstJob) goto L_ENTER_QUIT_PROCESS;
 		if (PSAgent.HasAccess()) goto L_ENTER_RUN_JOB;
 		UpdatePSPriority();
+		busy=true;
 		break;
 
 L_ENTER_QUIT_PROCESS:
-		CurrentDocument.Empty();
+		CurrentDocument.Clear();
 		PSAgent.ReleaseAccess();
 		Process.CloseWriting();
 		Process.CloseReading();
@@ -293,16 +299,13 @@ emPsRenderer::Job * emPsRenderer::SearchBestJob()
 emPsRenderer::Job * emPsRenderer::SearchBestSameDocJob()
 {
 	Job * job, * bestJob;
-	double bestPri;
 
 	for (bestJob=FirstJob; bestJob; bestJob=bestJob->Next) {
 		if (CurrentDocument==bestJob->Document) break;
 	}
 	if (bestJob) {
-		bestPri=bestJob->Priority;
 		for (job=bestJob->Next; job; job=job->Next) {
-			if (bestPri<job->Priority && CurrentDocument==job->Document) {
-				bestPri=job->Priority;
+			if (bestJob->Priority<job->Priority && bestJob->Document==job->Document) {
 				bestJob=job;
 			}
 		}
@@ -392,19 +395,28 @@ void emPsRenderer::UpdatePSPriority()
 }
 
 
-void emPsRenderer::TryStartProcess() throw(emString)
+void emPsRenderer::TryStartProcess()
 {
-#if defined(_WIN32)
-
-	throw emString(
-		"Ghostscript probably not available on this system. As a precaution, execution is not tried."
-	);
-
-#else
-
 	emArray<emString> args;
 
+#if defined(_WIN32)
+	args.Add(
+		emGetChildPath(
+			emGetInstallPath(EM_IDT_LIB,"emPs","emPs"),
+			"emPsWinAdapterProc"
+		)
+	);
+	const char * p=getenv("EM_DIR");
+	if (!p) throw emException("emPsRenderer: EM_DIR not set.");
+	emString gsPath=emString(p)+"\\thirdparty\\bin\\gs.exe";
+	if (!emIsRegularFile(gsPath)) {
+		// Otherwise we get a quite stupid error message.
+		throw emException("emPsRenderer: Cannot not find %s",gsPath.Get());
+	}
+	args.Add(gsPath);
+#else
 	args.Add("gs");
+#endif
 	args.Add("-q");
 	args.Add("-dNOPAUSE");
 	args.Add("-dSAFER");
@@ -424,10 +436,9 @@ void emPsRenderer::TryStartProcess() throw(emString)
 		NULL,
 		emProcess::SF_PIPE_STDIN|
 		emProcess::SF_PIPE_STDOUT|
-		emProcess::SF_SHARE_STDERR
+		emProcess::SF_SHARE_STDERR|
+		emProcess::SF_NO_WINDOW
 	);
-
-#endif
 }
 
 
@@ -435,7 +446,7 @@ void emPsRenderer::PrepareWritingStartup()
 {
 	WriterState=WRITING_STARTUP;
 	WriterPos=0;
-	WriteCommand.Empty();
+	WriteCommand.Clear();
 }
 
 
@@ -468,7 +479,7 @@ void emPsRenderer::PrepareWritingPage()
 }
 
 
-bool emPsRenderer::TryWrite() throw(emString)
+bool emPsRenderer::TryWrite()
 {
 	const char * buf;
 	int len;
@@ -515,7 +526,7 @@ L_ENTER_WRITING_SYNC:
 
 	len=Process.TryWrite(buf+WriterPos,len-WriterPos);
 	if (len<0) {
-		throw emString(
+		throw emException(
 			"PostScript interpretation failed: Interpreter closed STDIN or exited."
 		);
 	}
@@ -541,21 +552,21 @@ void emPsRenderer::PrepareReadingPage()
 }
 
 
-bool emPsRenderer::TryRead() throw(emString)
+bool emPsRenderer::TryRead()
 {
 	int len,syncLen,eat,r;
 	bool syncFound;
 	const char * p;
 
 	if (ReadBufferFill>=(int)sizeof(ReadBuffer)) {
-		throw emString("PostScript interpretation failed: Read buffer too small.");
+		throw emException("PostScript interpretation failed: Read buffer too small.");
 	}
 	len=Process.TryRead(
 		ReadBuffer+ReadBufferFill,
 		sizeof(ReadBuffer)-ReadBufferFill
 	);
 	if (len<0) {
-		throw emString(
+		throw emException(
 			"PostScript interpretation failed: Interpreter closed STDOUT or exited."
 		);
 	}
@@ -612,13 +623,13 @@ L_ENTER_READING_IMAGE_DATA:
 		while (len>0) {
 			r=ParseImageData(ReadBuffer+eat,len-eat);
 			if (r<0) {
-				throw emString(
+				throw emException(
 					"PostScript interpretation failed: Image data confusion."
 				);
 			}
 			else if (r==0) {
 				if (len-eat>1024) {
-					throw emString(
+					throw emException(
 						"PostScript interpretation failed: Image parser faulty."
 					);
 				}
@@ -649,7 +660,7 @@ L_ENTER_READING_SYNC:
 	}
 
 	if (syncFound && ReaderState!=READING_FINISHED) {
-		throw emString(
+		throw emException(
 			"PostScript interpretation failed: Unsupported document structure."
 		);
 	}
