@@ -1,7 +1,7 @@
 //------------------------------------------------------------------------------
 // emWndsWindowPort.cpp
 //
-// Copyright (C) 2006-2010 Oliver Hamann.
+// Copyright (C) 2006-2012,2014-2017 Oliver Hamann.
 //
 // Homepage: http://eaglemode.sourceforge.net/
 //
@@ -19,49 +19,128 @@
 //------------------------------------------------------------------------------
 
 #include <emWnds/emWndsWindowPort.h>
+#include <emWnds/emWndsViewRenderer.h>
+
+#ifndef WM_UNICHAR
+#	define WM_UNICHAR 0x0109
+#endif
 
 #ifndef WM_MOUSEWHEEL
 #	define WM_MOUSEWHEEL 0x020A
 #endif
-#ifndef WM_UNICHAR
-#	define WM_UNICHAR 0x0109
+
+#ifndef WM_MOUSEHWHEEL
+#	define WM_MOUSEHWHEEL 0x020E
+#endif
+
+#ifndef WM_XBUTTONDOWN
+#	define WM_XBUTTONDOWN 0x020B
+#endif
+
+#ifndef WM_XBUTTONUP
+#	define WM_XBUTTONUP 0x020C
+#endif
+
+#ifndef WM_XBUTTONDBLCLK
+#	define WM_XBUTTONDBLCLK 0x020D
+#endif
+
+#ifndef MK_XBUTTON1
+#	define MK_XBUTTON1 0x0020
+#endif
+
+#ifndef MK_XBUTTON2
+#	define MK_XBUTTON2 0x0040
+#endif
+
+#ifndef XBUTTON1
+#	define XBUTTON1 0x0001
+#endif
+
+#ifndef XBUTTON2
+#	define XBUTTON2 0x0002
 #endif
 
 
 void emWndsWindowPort::WindowFlagsChanged()
 {
+	emWindow::WindowFlags oldFlags;
 	HWND oldHWnd;
 	int i;
 
-	oldHWnd=HWnd;
-	HWnd=NULL;
+	oldFlags=WindowFlags;
+	WindowFlags=GetWindowFlags();
 
-	SetModalState(false);
+	if (
+		((oldFlags^WindowFlags) & (
+			emWindow::WF_MODAL |
+			emWindow::WF_UNDECORATED |
+			emWindow::WF_POPUP |
+			emWindow::WF_FULLSCREEN
+		)) != 0
+	) {
+		oldHWnd=HWnd;
+		HWnd=NULL;
 
-	PreConstruct();
+		SetModalState(false);
 
-	for (i=0; i<Screen.WinPorts.GetCount(); i++) {
-		if (
-			Screen.WinPorts[i]->Owner==this &&
-			Screen.WinPorts[i]->HWnd!=NULL
-		) {
-			// It is said that SetParent(...) must be used instead of
-			// SetWindowLong(...GWL_HWNDPARENT...). But with SetParent the
-			// window behaves like WS_CHILD always - even when playing
-			// with ShowWindow and SetWindowLong(GWL_STYLE/GWL_EXSTYLE).
-			// Another problem is: It does not work if the owner window
-			// has WS_EX_TOPMOST (even BringWindowToTop does not help here).
-#			if defined(_WIN64)
-				SetWindowLongPtr(Screen.WinPorts[i]->HWnd,GWLP_HWNDPARENT,(LONG_PTR)HWnd);
-#			else
-				SetWindowLong(Screen.WinPorts[i]->HWnd,GWL_HWNDPARENT,(LONG)HWnd);
-#			endif
+		PreConstruct();
+
+		for (i=0; i<Screen.WinPorts.GetCount(); i++) {
+			if (
+				Screen.WinPorts[i]->Owner==this &&
+				Screen.WinPorts[i]->HWnd!=NULL
+			) {
+				// It is said that SetParent(...) must be used instead of
+				// SetWindowLong(...GWL_HWNDPARENT...). But with SetParent the
+				// window behaves like WS_CHILD always - even when playing
+				// with ShowWindow and SetWindowLong(GWL_STYLE/GWL_EXSTYLE).
+				// Another problem is: It does not work if the owner window
+				// has WS_EX_TOPMOST (even BringWindowToTop does not help here).
+#				if defined(_WIN64)
+					SetWindowLongPtr(Screen.WinPorts[i]->HWnd,GWLP_HWNDPARENT,(LONG_PTR)HWnd);
+#				else
+					SetWindowLong(Screen.WinPorts[i]->HWnd,GWL_HWNDPARENT,(LONG)HWnd);
+#				endif
+			}
 		}
+
+		DestroyWindow(oldHWnd);
+		if (BigIcon) { DestroyIcon(BigIcon); BigIcon=NULL; }
+		if (SmallIcon) { DestroyIcon(SmallIcon); SmallIcon=NULL; }
+
+		if (
+			(oldFlags&emWindow::WF_FULLSCREEN)!=0 &&
+			(WindowFlags&emWindow::WF_FULLSCREEN)==0 &&
+			RestoreW>0 && RestoreH>0
+		) {
+			SetViewGeometry(RestoreX,RestoreY,RestoreW,RestoreH,Screen.PixelTallness);
+			PosForced=true;
+			PosPending=true;
+			SizeForced=true;
+			SizePending=true;
+			WakeUp();
+		}
+
+		return;
 	}
 
-	DestroyWindow(oldHWnd);
-	if (BigIcon) { DestroyIcon(BigIcon); BigIcon=NULL; }
-	if (SmallIcon) { DestroyIcon(SmallIcon); SmallIcon=NULL; }
+	if (PostConstructed) {
+		if (
+			((oldFlags^WindowFlags) & emWindow::WF_MAXIMIZED) != 0 &&
+			(WindowFlags&(
+				emWindow::WF_UNDECORATED|
+				emWindow::WF_POPUP|
+				emWindow::WF_FULLSCREEN
+			))==0
+		) {
+			ShowWindowAsync(
+				HWnd,
+				(WindowFlags&emWindow::WF_MAXIMIZED)!=0 ?
+				SW_MAXIMIZE : SW_RESTORE
+			);
+		}
+	}
 }
 
 
@@ -73,7 +152,7 @@ void emWndsWindowPort::SetPosSize(
 	RECT rw,rc;
 	POINT pnt;
 
-	if ((GetWindowFlags()&emWindow::WF_FULLSCREEN)!=0) {
+	if ((WindowFlags&(emWindow::WF_MAXIMIZED|emWindow::WF_FULLSCREEN))!=0) {
 		posSpec=PSAS_IGNORE;
 		sizeSpec=PSAS_IGNORE;
 	}
@@ -113,6 +192,12 @@ void emWndsWindowPort::SetPosSize(
 		SizeForced=true;
 		SizePending=true;
 	}
+	if ((WindowFlags&(emWindow::WF_MAXIMIZED|emWindow::WF_FULLSCREEN))==0) {
+		RestoreX=x;
+		RestoreY=y;
+		RestoreW=w;
+		RestoreH=h;
+	}
 	SetViewGeometry(x,y,w,h,Screen.PixelTallness);
 	WakeUp();
 }
@@ -120,7 +205,7 @@ void emWndsWindowPort::SetPosSize(
 
 void emWndsWindowPort::GetBorderSizes(
 	double * pL, double * pT, double * pR, double * pB
-)
+) const
 {
 	RECT rw,rc;
 	POINT pnt;
@@ -155,6 +240,24 @@ void emWndsWindowPort::Raise()
 }
 
 
+void emWndsWindowPort::InhibitScreensaver()
+{
+	ScreensaverInhibitCount++;
+}
+
+
+void emWndsWindowPort::AllowScreensaver()
+{
+	ScreensaverInhibitCount--;
+}
+
+
+emUInt64 emWndsWindowPort::GetInputClockMS() const
+{
+	return emGetClockMS(); // ???
+}
+
+
 void emWndsWindowPort::InvalidateTitle()
 {
 	TitlePending=true;
@@ -188,7 +291,8 @@ void emWndsWindowPort::InvalidatePainting(double x, double y, double w, double h
 	if (y2>ClipY2) y2=ClipY2;
 	if (y<ClipY1) y=ClipY1;
 	if (y>=y2) return;
-	MergeToInvRectList((int)x,(int)y,(int)ceil(x2),(int)ceil(y2));
+	InvalidRects.Unite((int)x,(int)y,(int)ceil(x2),(int)ceil(y2));
+	if (InvalidRects.GetCount()>64) InvalidRects.SetToMinMax();
 	WakeUp();
 }
 
@@ -201,7 +305,6 @@ emWndsWindowPort::emWndsWindowPort(emWindow & window)
 	emContext * c;
 	emWndsWindowPort * wp;
 	emWindow * w;
-	int i;
 
 	Owner=NULL;
 	for (c=window.GetParentContext(); c; c=c->GetParentContext()) {
@@ -214,12 +317,17 @@ emWndsWindowPort::emWndsWindowPort(emWindow & window)
 		break;
 	}
 	HWnd=NULL;
+	WindowFlags=GetWindowFlags();
 	MinPaneW=1;
 	MinPaneH=1;
 	PaneX=0;
 	PaneY=0;
 	PaneW=1;
 	PaneH=1;
+	RestoreX=-1;
+	RestoreY=-1;
+	RestoreW=-1;
+	RestoreH=-1;
 	ClipX1=PaneX;
 	ClipY1=PaneY;
 	ClipX2=PaneX+PaneW;
@@ -235,12 +343,6 @@ emWndsWindowPort::emWndsWindowPort(emWindow & window)
 	TitlePending=false;
 	IconPending=false;
 	CursorPending=false;
-	InvRectFreeList=NULL;
-	InvRectList=NULL;
-	for (i=0; i<(int)(sizeof(InvRectHeap)/sizeof(InvRect)); i++) {
-		InvRectHeap[i].Next=InvRectFreeList;
-		InvRectFreeList=&InvRectHeap[i];
-	}
 	InputStateClock=0;
 	LastButtonPress=EM_KEY_NONE;
 	LastButtonPressTime=0;
@@ -250,10 +352,12 @@ emWndsWindowPort::emWndsWindowPort(emWindow & window)
 	RepeatKey=EM_KEY_NONE;
 	KeyRepeat=0;
 	MouseWheelRemainder=0;
+	MouseHWheelRemainder=0;
 	ModalState=false;
 	ModalDescendants=0;
 	BigIcon=NULL;
 	SmallIcon=NULL;
+	ScreensaverInhibitCount=0;
 
 	Screen.WinPorts.Add(this);
 
@@ -284,12 +388,18 @@ void emWndsWindowPort::PreConstruct()
 {
 	double vrx,vry,vrw,vrh,d;
 	DWORD exStyleFlags,styleFlags;
-	int x,y,w,h;
+	int x,y,w,h,monitor;
 	bool haveBorder;
 
-	if ((GetWindowFlags()&emWindow::WF_FULLSCREEN)!=0) {
-		Screen.LeaveFullscreenModes(&GetWindow());
-		Screen.GetVisibleRect(&vrx,&vry,&vrw,&vrh);
+	monitor=0;
+	if (Owner) monitor=Owner->GetWindow().GetMonitorIndex();
+	Screen.GetMonitorRect(monitor,&vrx,&vry,&vrw,&vrh);
+
+	if ((WindowFlags&emWindow::WF_FULLSCREEN)!=0) {
+		if (RestoreW>0 && RestoreH>0) {
+			monitor=Screen.GetMonitorIndexOfRect(RestoreX,RestoreY,RestoreW,RestoreH);
+			Screen.GetMonitorRect(monitor,&vrx,&vry,&vrw,&vrh);
+		}
 		MinPaneW=1;
 		MinPaneH=1;
 		PaneX=(int)(vrx+0.5);
@@ -298,8 +408,7 @@ void emWndsWindowPort::PreConstruct()
 		PaneH=(int)(vrh+0.5);
 		haveBorder=false;
 	}
-	else if ((GetWindowFlags()&(emWindow::WF_POPUP|emWindow::WF_UNDECORATED))!=0) {
-		Screen.GetVisibleRect(&vrx,&vry,&vrw,&vrh);
+	else if ((WindowFlags&(emWindow::WF_POPUP|emWindow::WF_UNDECORATED))!=0) {
 		MinPaneW=1;
 		MinPaneH=1;
 		PaneX=(int)(vrx+vrw*emGetDblRandom(0.22,0.28)+0.5);
@@ -309,11 +418,9 @@ void emWndsWindowPort::PreConstruct()
 		haveBorder=false;
 	}
 	else {
-		if (!Owner) Screen.LeaveFullscreenModes(&GetWindow());
-		Screen.GetVisibleRect(&vrx,&vry,&vrw,&vrh);
 		MinPaneW=32;
 		MinPaneH=32;
-		if (!Owner && (GetWindowFlags()&emWindow::WF_MODAL)==0) {
+		if (!Owner && (WindowFlags&emWindow::WF_MODAL)==0) {
 			d=emMin(vrw,vrh)*0.08;
 			PaneX=(int)(vrx+d*emGetDblRandom(0.5,1.5)+0.5);
 			PaneY=(int)(vry+d*emGetDblRandom(0.8,1.2)+0.5);
@@ -337,9 +444,8 @@ void emWndsWindowPort::PreConstruct()
 	PosPending=false;
 	SizeForced=false;
 	SizePending=false;
-	ClearInvRectList();
-	MergeToInvRectList(PaneX,PaneY,PaneX+PaneW,PaneY+PaneH);
-	Title.Empty();
+	InvalidRects.Set(PaneX,PaneY,PaneX+PaneW,PaneY+PaneH);
+	Title.Clear();
 	TitlePending=true;
 	IconPending=true;
 	Cursor=-1;
@@ -352,13 +458,14 @@ void emWndsWindowPort::PreConstruct()
 	LastButtonPress=EM_KEY_NONE;
 	RepeatKey=EM_KEY_NONE;
 	MouseWheelRemainder=0;
+	MouseHWheelRemainder=0;
 
 	if (haveBorder) {
 		if (Owner) {
 			exStyleFlags=0;
 			styleFlags=WS_OVERLAPPED|WS_CAPTION|WS_SYSMENU|WS_THICKFRAME|WS_MAXIMIZEBOX;
-			x=CW_USEDEFAULT;
-			y=CW_USEDEFAULT;
+			x=PaneX;
+			y=PaneY;
 			w=PaneW;
 			h=PaneH;
 		}
@@ -423,9 +530,24 @@ void emWndsWindowPort::PreConstruct()
 
 void emWndsWindowPort::PostConstruct()
 {
-	ShowWindow(HWnd,SW_SHOWNORMAL);
 
-	if ((GetWindowFlags()&emWindow::WF_MODAL)!=0) {
+	if (
+		(WindowFlags&emWindow::WF_MAXIMIZED)!=0 &&
+		(WindowFlags&(
+			emWindow::WF_UNDECORATED|
+			emWindow::WF_POPUP|
+			emWindow::WF_FULLSCREEN
+		))==0
+	) {
+		ShowWindow(HWnd,SW_SHOWMAXIMIZED);
+	}
+	else {
+		ShowWindow(HWnd,SW_SHOWNORMAL);
+	}
+
+	Mapped=true;
+
+	if ((WindowFlags&emWindow::WF_MODAL)!=0) {
 		SetModalState(true);
 	}
 }
@@ -460,9 +582,11 @@ LRESULT emWndsWindowPort::WindowProc(
 	case WM_LBUTTONDOWN:
 	case WM_MBUTTONDOWN:
 	case WM_RBUTTONDOWN:
+	case WM_XBUTTONDOWN:
 	case WM_LBUTTONDBLCLK:
 	case WM_MBUTTONDBLCLK:
 	case WM_RBUTTONDBLCLK:
+	case WM_XBUTTONDBLCLK:
 		mx=PaneX+((short)LOWORD(lParam))+Screen.MouseWarpX;
 		my=PaneY+((short)HIWORD(lParam))+Screen.MouseWarpY;
 		if (
@@ -473,19 +597,54 @@ LRESULT emWndsWindowPort::WindowProc(
 			Screen.InputStateClock++;
 		}
 		switch (uMsg) {
-		case WM_LBUTTONDOWN  : key=EM_KEY_LEFT_BUTTON  ; dblClick=false; break;
-		case WM_MBUTTONDOWN  : key=EM_KEY_MIDDLE_BUTTON; dblClick=false; break;
-		case WM_RBUTTONDOWN  : key=EM_KEY_RIGHT_BUTTON ; dblClick=false; break;
-		case WM_LBUTTONDBLCLK: key=EM_KEY_LEFT_BUTTON  ; dblClick=true ; break;
-		case WM_MBUTTONDBLCLK: key=EM_KEY_MIDDLE_BUTTON; dblClick=true ; break;
-		case WM_RBUTTONDBLCLK: key=EM_KEY_RIGHT_BUTTON ; dblClick=true ; break;
-		default              : key=EM_KEY_NONE         ; dblClick=false; break;
+		case WM_LBUTTONDOWN:
+			key=EM_KEY_LEFT_BUTTON;
+			dblClick=false;
+			break;
+		case WM_MBUTTONDOWN:
+			key=EM_KEY_MIDDLE_BUTTON;
+			dblClick=false;
+			break;
+		case WM_RBUTTONDOWN:
+			key=EM_KEY_RIGHT_BUTTON;
+			dblClick=false;
+			break;
+		case WM_XBUTTONDOWN:
+			if (HIWORD(wParam)==XBUTTON1) key=EM_KEY_BACK_BUTTON;
+			else if (HIWORD(wParam)==XBUTTON2) key=EM_KEY_FORWARD_BUTTON;
+			else key=EM_KEY_NONE;
+			dblClick=false;
+			break;
+		case WM_LBUTTONDBLCLK:
+			key=EM_KEY_LEFT_BUTTON;
+			dblClick=true;
+			break;
+		case WM_MBUTTONDBLCLK:
+			key=EM_KEY_MIDDLE_BUTTON;
+			dblClick=true;
+			break;
+		case WM_RBUTTONDBLCLK:
+			key=EM_KEY_RIGHT_BUTTON;
+			dblClick=true;
+			break;
+		case WM_XBUTTONDBLCLK:
+			if (HIWORD(wParam)==XBUTTON1) key=EM_KEY_BACK_BUTTON;
+			else if (HIWORD(wParam)==XBUTTON2) key=EM_KEY_FORWARD_BUTTON;
+			else key=EM_KEY_NONE;
+			dblClick=true;
+			break;
+		default:
+			key=EM_KEY_NONE;
+			dblClick=false;
+			break;
 		}
 		if (key==EM_KEY_NONE) {
-			for (i=0; i<3; i++) {
-				if      (i==0) { key=EM_KEY_LEFT_BUTTON  ; mask=MK_LBUTTON; }
-				else if (i==1) { key=EM_KEY_MIDDLE_BUTTON; mask=MK_MBUTTON; }
-				else           { key=EM_KEY_RIGHT_BUTTON ; mask=MK_RBUTTON; }
+			for (i=0; i<5; i++) {
+				if      (i==0) { key=EM_KEY_LEFT_BUTTON   ; mask=MK_LBUTTON;  }
+				else if (i==1) { key=EM_KEY_MIDDLE_BUTTON ; mask=MK_MBUTTON;  }
+				else if (i==2) { key=EM_KEY_RIGHT_BUTTON  ; mask=MK_RBUTTON;  }
+				else if (i==3) { key=EM_KEY_BACK_BUTTON   ; mask=MK_XBUTTON1; }
+				else           { key=EM_KEY_FORWARD_BUTTON; mask=MK_XBUTTON2; }
 				if (Screen.InputState.Get(key) && (wParam&mask)==0) {
 					Screen.InputState.Set(key,false);
 					Screen.InputStateClock++;
@@ -550,6 +709,7 @@ LRESULT emWndsWindowPort::WindowProc(
 	case WM_LBUTTONUP:
 	case WM_MBUTTONUP:
 	case WM_RBUTTONUP:
+	case WM_XBUTTONUP:
 		mx=PaneX+((short)LOWORD(lParam))+Screen.MouseWarpX;
 		my=PaneY+((short)HIWORD(lParam))+Screen.MouseWarpY;
 		if (
@@ -559,9 +719,15 @@ LRESULT emWndsWindowPort::WindowProc(
 			Screen.InputState.SetMouse(mx,my);
 			Screen.InputStateClock++;
 		}
-		if (uMsg==WM_LBUTTONUP)      key=EM_KEY_LEFT_BUTTON;
-		else if (uMsg==WM_MBUTTONUP) key=EM_KEY_MIDDLE_BUTTON;
-		else                         key=EM_KEY_RIGHT_BUTTON;
+		if (uMsg==WM_LBUTTONUP)            key=EM_KEY_LEFT_BUTTON;
+		else if (uMsg==WM_MBUTTONUP)       key=EM_KEY_MIDDLE_BUTTON;
+		else if (uMsg==WM_RBUTTONUP)       key=EM_KEY_RIGHT_BUTTON;
+		else if (HIWORD(wParam)==XBUTTON1) key=EM_KEY_BACK_BUTTON;
+		else if (HIWORD(wParam)==XBUTTON2) key=EM_KEY_FORWARD_BUTTON;
+		else                               key=EM_KEY_NONE;
+		if (key==EM_KEY_NONE) {
+			return 0;
+		}
 		if (Screen.InputState.Get(key)) {
 			Screen.InputState.Set(key,false);
 			Screen.InputStateClock++;
@@ -574,7 +740,9 @@ LRESULT emWndsWindowPort::WindowProc(
 			GetCapture()==HWnd &&
 			!Screen.InputState.Get(EM_KEY_LEFT_BUTTON) &&
 			!Screen.InputState.Get(EM_KEY_MIDDLE_BUTTON) &&
-			!Screen.InputState.Get(EM_KEY_RIGHT_BUTTON)
+			!Screen.InputState.Get(EM_KEY_RIGHT_BUTTON) &&
+			!Screen.InputState.Get(EM_KEY_BACK_BUTTON) &&
+			!Screen.InputState.Get(EM_KEY_FORWARD_BUTTON)
 		) {
 			ReleaseCapture();
 		}
@@ -586,6 +754,7 @@ LRESULT emWndsWindowPort::WindowProc(
 		}
 		return 0;
 	case WM_MOUSEWHEEL:
+	case WM_MOUSEHWHEEL:
 		mx=((short)LOWORD(lParam))+Screen.MouseWarpX;
 		my=((short)HIWORD(lParam))+Screen.MouseWarpY;
 		if (
@@ -621,29 +790,39 @@ LRESULT emWndsWindowPort::WindowProc(
 					Screen.UpdateKeymapAndInputState();
 				}
 				if (mx>=PaneX && mx<PaneX+PaneW && my>=PaneY && my<PaneY+PaneH) {
-					MouseWheelRemainder+=(short)HIWORD(wParam);
-					y=MouseWheelRemainder/120;
-					MouseWheelRemainder-=y*120;
-					if (y!=0) {
-						if (y>0) key=EM_KEY_WHEEL_UP;
-						else key=EM_KEY_WHEEL_DOWN;
-						inputEvent.Setup(key,"",0,0);
-						InputStateClock=Screen.InputStateClock;
-						InputToView(inputEvent,Screen.InputState);
-						return 0;
+					if (uMsg==WM_MOUSEWHEEL) {
+						MouseWheelRemainder+=(short)HIWORD(wParam);
+						y=MouseWheelRemainder/120;
+						MouseWheelRemainder-=y*120;
+						if (y!=0) {
+							if (y>0) key=EM_KEY_WHEEL_UP;
+							else key=EM_KEY_WHEEL_DOWN;
+							inputEvent.Setup(key,"",0,0);
+							InputStateClock=Screen.InputStateClock;
+							InputToView(inputEvent,Screen.InputState);
+							return 0;
+						}
+					}
+					else {
+						MouseHWheelRemainder+=(short)HIWORD(wParam);
+						x=MouseHWheelRemainder/120;
+						MouseHWheelRemainder-=x*120;
+						if (x!=0) {
+							if (x>0) key=EM_KEY_WHEEL_RIGHT;
+							else key=EM_KEY_WHEEL_LEFT;
+							inputEvent.Setup(key,"",0,0);
+							InputStateClock=Screen.InputStateClock;
+							InputToView(inputEvent,Screen.InputState);
+							return 0;
+						}
 					}
 				}
 			}
 		}
 		else {
 			// Here we have a mouse wheel event while the mouse is in
-			// another window. And we transfer the focus and the mouse
-			// wheel event to that window, even if it is a foreign
-			// application. This is highly incompatible to the normal
-			// mouse wheel behavior in Windows.
+			// another window. Simply forward the message.
 			if (IsWindowEnabled(hwnd)) {
-				if (wp) SetActiveWindow(hwnd);
-				else SetForegroundWindow(hwnd);
 				PostMessage(hwnd,uMsg,wParam,lParam);
 			}
 		}
@@ -695,6 +874,11 @@ LRESULT emWndsWindowPort::WindowProc(
 	case WM_SYSDEADCHAR:
 	case WM_UNICHAR:
 		return 0;
+	case WM_SYSCOMMAND:
+		if (wParam==SC_SCREENSAVE || wParam==SC_MONITORPOWER) {
+			if (IsScreensaverReallyToInhibit()) return 0;
+		}
+		return DefWindowProc(HWnd,uMsg,wParam,lParam);
 	case WM_PAINT:
 		hdc=BeginPaint(HWnd,&ps);
 		if (hdc) {
@@ -702,14 +886,7 @@ LRESULT emWndsWindowPort::WindowProc(
 			y=ps.rcPaint.top;
 			w=ps.rcPaint.right-ps.rcPaint.left;
 			h=ps.rcPaint.bottom-ps.rcPaint.top;
-			if (x<0) { w+=x; x=0; }
-			if (y<0) { h+=y; y=0; }
-			if (w>PaneW-x) w=PaneW-x;
-			if (h>PaneH-y) h=PaneH-y;
-			if (w>0 && h>0) {
-				MergeToInvRectList(PaneX+x,PaneY+y,PaneX+x+w,PaneY+y+h);
-				WakeUp();
-			}
+			InvalidatePainting(PaneX+x,PaneY+y,w,h);
 			EndPaint(HWnd,&ps);
 		}
 		return 0;
@@ -726,22 +903,42 @@ LRESULT emWndsWindowPort::WindowProc(
 		Screen.UpdateKeymapAndInputState();
 		RepeatKey=EM_KEY_NONE;
 		MouseWheelRemainder=0;
+		MouseHWheelRemainder=0;
 		if (!Focused) {
 			Focused=true;
 			SetViewFocused(true);
 		}
+		for (i=Screen.WinPorts.GetCount()-1; i>=0; i--) {
+			wp=Screen.WinPorts[i];
+			if (
+				(wp->GetWindowFlags()&emWindow::WF_POPUP)!=0 &&
+				wp!=this && !wp->IsAncestorOf(this)
+			) {
+				wp->SignalWindowClosing();
+			}
+		}
 		return 0;
 	case WM_KILLFOCUS:
 		if (GetCapture()==HWnd) ReleaseCapture();
 		if (Focused) {
 			Focused=false;
 			SetViewFocused(false);
-			if ((GetWindowFlags()&emWindow::WF_POPUP)!=0) SignalWindowClosing();
 		}
 		LastButtonPress=EM_KEY_NONE;
 		RepeatKey=EM_KEY_NONE;
 		MouseWheelRemainder=0;
+		MouseHWheelRemainder=0;
 		return 0;
+	case WM_ACTIVATEAPP:
+		if (wParam==false) {
+			for (i=Screen.WinPorts.GetCount()-1; i>=0; i--) {
+				wp=Screen.WinPorts[i];
+				if ((wp->GetWindowFlags()&emWindow::WF_POPUP)!=0) {
+					wp->SignalWindowClosing();
+				}
+			}
+		}
+		return DefWindowProc(HWnd,uMsg,wParam,lParam);
 	case WM_SETCURSOR:
 		if ((HWND)wParam!=HWnd || LOWORD(lParam)!=1) {
 			CursorShown=false;
@@ -770,8 +967,7 @@ LRESULT emWndsWindowPort::WindowProc(
 			ClipY1=PaneY;
 			ClipX2=PaneX+PaneW;
 			ClipY2=PaneY+PaneH;
-			ClearInvRectList();
-			MergeToInvRectList(PaneX,PaneY,PaneX+PaneW,PaneY+PaneH);
+			InvalidRects.Set(PaneX,PaneY,PaneX+PaneW,PaneY+PaneH);
 			WakeUp();
 			if (!PosPending && !SizePending) {
 				SetViewGeometry(
@@ -796,7 +992,39 @@ LRESULT emWndsWindowPort::WindowProc(
 			}
 			Screen.InputStateClock++;
 		}
+		DefWindowProc(HWnd,uMsg,wParam,lParam); // Calls WM_SIZE if needed.
+		if ((WindowFlags&(emWindow::WF_MAXIMIZED|emWindow::WF_FULLSCREEN))==0) {
+			RestoreX=PaneX;
+			RestoreY=PaneY;
+			RestoreW=PaneW;
+			RestoreH=PaneH;
+		}
 		return 0;
+	case WM_SIZE:
+		if (PostConstructed) {
+			if (wParam==SIZE_MINIMIZED) {
+				Mapped=false;
+			}
+			else {
+				if (!Mapped) {
+					Mapped=true;
+					WakeUp();
+				}
+				if (wParam==SIZE_MAXIMIZED) {
+					if ((WindowFlags&emWindow::WF_MAXIMIZED)==0) {
+						WindowFlags|=emWindow::WF_MAXIMIZED;
+						GetWindow().SetWindowFlags(WindowFlags);
+					}
+				}
+				else if (wParam==SIZE_RESTORED) {
+					if ((WindowFlags&emWindow::WF_MAXIMIZED)!=0) {
+						WindowFlags&=~emWindow::WF_MAXIMIZED;
+						GetWindow().SetWindowFlags(WindowFlags);
+					}
+				}
+			}
+		}
+		return DefWindowProc(HWnd,uMsg,wParam,lParam);
 	case WM_GETMINMAXINFO:
 		DefWindowProc(HWnd,uMsg,wParam,lParam);
 		GetClientRect(HWnd,&rc);
@@ -807,19 +1035,6 @@ LRESULT emWndsWindowPort::WindowProc(
 		if (pPnt->x<w) pPnt->x=w;
 		if (pPnt->y<h) pPnt->y=h;
 		return 0;
-	case WM_SHOWWINDOW:
-		if (wParam) {
-			if (!Mapped) {
-				Mapped=true;
-				WakeUp();
-			}
-		}
-		else {
-			if (Mapped) {
-				Mapped=false;
-			}
-		}
-		return DefWindowProc(HWnd,uMsg,wParam,lParam);
 	case WM_CLOSE:
 		SignalWindowClosing();
 		return 0;
@@ -912,7 +1127,7 @@ bool emWndsWindowPort::Cycle()
 		PostConstructed=true;
 	}
 
-	if (InvRectList && Mapped) UpdatePainting();
+	if (!InvalidRects.IsEmpty() && Mapped) UpdatePainting();
 
 	return false;
 }
@@ -920,138 +1135,43 @@ bool emWndsWindowPort::Cycle()
 
 void emWndsWindowPort::UpdatePainting()
 {
-	InvRect * r;
-	HDC hdc;
-	int bufIdx,rx1,ry1,rx2,ry2,x,y,w,h;
+	InvalidRects.Sort();
 
-	if (InvRectList) {
-		hdc=GetDC(HWnd);
-		bufIdx=0;
-		do {
-			r=InvRectList;
-			rx1=r->x1;
-			ry1=r->y1;
-			rx2=r->x2;
-			ry2=r->y2;
-			InvRectList=r->Next;
-			r->Next=InvRectFreeList;
-			InvRectFreeList=r;
-			y=ry1;
-			do {
-				h=ry2-y;
-				if (h>Screen.BufHeight) h=Screen.BufHeight;
-				x=rx1;
-				do {
-					w=rx2-x;
-					if (w>Screen.BufWidth) w=Screen.BufWidth;
-					PaintView(emPainter(Screen.BufPainter[bufIdx],0,0,w,h,-x,-y,1,1),0);
-					Screen.BeginSendBuf(hdc,bufIdx,x-PaneX,y-PaneY,w,h);
-					bufIdx^=1;
-					x+=w;
-				} while (x<rx2);
-				y+=h;
-			} while (y<ry2);
-		} while (InvRectList);
-		Screen.WaitSendBuf();
-	}
+	Screen.ViewRenderer->RenderView(
+		*this,
+		InvalidRects,
+		GetDC(HWnd)
+	);
+
+	InvalidRects.Clear();
 }
 
 
-void emWndsWindowPort::ClearInvRectList()
+bool emWndsWindowPort::IsScreensaverReallyToInhibit()
 {
-	InvRect * r;
+	double x1,y1,x2,y2,vx,vy,vw,vh,mx,my,mw,mh;
+	int i,n;
 
-	while ((r=InvRectList)!=NULL) {
-		InvRectList=r->Next;
-		r->Next=InvRectFreeList;
-		InvRectFreeList=r;
-	}
-}
+	if (ScreensaverInhibitCount<=0) return false;
 
+	if (::IsIconic(HWnd)) return false;
 
-void emWndsWindowPort::MergeToInvRectList(int x1, int y1, int x2, int y2)
-{
-	InvRect * * pr;
-	InvRect * r;
-	int rx1,ry1,rx2,ry2;
-
-	pr=&InvRectList;
-	for (;;) {
-		r=*pr;
-		if (!r || (ry1=r->y1)>y2) break;
-		if ((ry2=r->y2)<y1 || (rx1=r->x1)>x2 || (rx2=r->x2)<x1) {
-			pr=&r->Next;
-		}
-		else if (rx1>=x1 && rx2<=x2 && ry1>=y1 && ry2<=y2) {
-			*pr=r->Next;
-			r->Next=InvRectFreeList;
-			InvRectFreeList=r;
-		}
-		else if (rx1<=x1 && rx2>=x2 && ry1<=y1 && ry2>=y2) {
-			return;
-		}
-		else if (rx1==x1 && rx2==x2) {
-			if (y1>ry1) y1=ry1;
-			if (y2<ry2) y2=ry2;
-			*pr=r->Next;
-			r->Next=InvRectFreeList;
-			InvRectFreeList=r;
-			pr=&InvRectList;
-		}
-		else if (ry1<y2 && ry2>y1) {
-			*pr=r->Next;
-			r->Next=InvRectFreeList;
-			InvRectFreeList=r;
-			if (ry1<y1) {
-				MergeToInvRectList(rx1,ry1,rx2,y1);
-			}
-			else if (y1<ry1) {
-				MergeToInvRectList(x1,y1,x2,ry1);
-				y1=ry1;
-			}
-			if (ry2>y2) {
-				MergeToInvRectList(rx1,y2,rx2,ry2);
-			}
-			else if (ry2<y2) {
-				MergeToInvRectList(x1,ry2,x2,y2);
-				y2=ry2;
-			}
-			if (x1>rx1) x1=rx1;
-			if (x2<rx2) x2=rx2;
-			pr=&InvRectList;
-		}
-		else {
-			pr=&r->Next;
+	vx=GetViewX();
+	vy=GetViewY();
+	vw=GetViewWidth();
+	vh=GetViewHeight();
+	n=Screen.GetMonitorCount();
+	for (i=0; i<n; i++) {
+		Screen.GetMonitorRect(i,&mx,&my,&mw,&mh);
+		x1=emMax(vx,mx);
+		y1=emMax(vy,my);
+		x2=emMin(vx+vw,mx+mw);
+		y2=emMin(vy+vh,my+mh);
+		if (x1<x2 && y1<y2 && (x2-x1)*(y2-y1)>=0.6*mw*mh) {
+			return true;
 		}
 	}
-
-	pr=&InvRectList;
-	for (;;) {
-		r=*pr;
-		if (!r || r->y1>y1 || (r->y1==y1 && r->x1>x1)) break;
-		pr=&r->Next;
-	}
-	r=InvRectFreeList;
-	if (!r) {
-		while ((r=InvRectList)!=NULL) {
-			if (x1>r->x1) x1=r->x1;
-			if (x2<r->x2) x2=r->x2;
-			if (y1>r->y1) y1=r->y1;
-			if (y2<r->y2) y2=r->y2;
-			InvRectList=r->Next;
-			r->Next=InvRectFreeList;
-			InvRectFreeList=r;
-		}
-		pr=&InvRectList;
-		r=InvRectFreeList;
-	}
-	InvRectFreeList=r->Next;
-	r->x1=x1;
-	r->y1=y1;
-	r->x2=x2;
-	r->y2=y2;
-	r->Next=*pr;
-	*pr=r;
+	return false;
 }
 
 
